/* simple program for BrickOS (version 0.9) that sends an IR message from RCX to WowWee robots */

/* PC -> RCX protocol (LNP integrity messages):
 *
 *  all messages are 3-byte
 *
 *  immediate action:
 *    0x7D b1 b2      (where b1 is LSB of hexcode, and b2 is MSB of hexcode)
 *
 *  program a sequence:
 *    0x7E b1 b2      start of message (b1, b2 unused)
 *    0x7D b1 b2      hexcode for action (as above) - the time when the message is sent is recorded for replay
 *    0x7A b1 b2      overwrite the time with a specified value (b1 is LSB), value in milliseconds (7A should follow after 7D)
 *    0x7C b1 b2      end of message, replay immediatelly, or:
 *    0x7B b1 b2      end of message, store it, and replay when PRGM button is pressed (even multiple times)
 *
 */

#include <conio.h>
#include <unistd.h>
#include <dmotor.h>
#include <dbutton.h>
#include <lnp/lnp.h>
#include <lnp/lnp-logical.h>

#include <sys/h8.h>
#include <sys/time.h>

inline void wait_unit()
{
  unsigned char b, c;

  for (b = 0; b < 255; b++)
    for (c = 0; c < 1; c++);
  for (b = 0; b < 25; b++)
    for (c = 0; c < 1; c++);
}

void sapien_send(int code)
{
  int x = (code < 0x100)?128:2048;

  asm("ORC #128, ccr");  /* clear interrupts */

  /* clear TE and RE bits */
  S_CR = S_CR & 0xCF;

  /* set direction on the P50 (TxD) to out */
  PORT5_DDR = 5;

  /* header */

  PORT5 = 0;
  /* wait 8/1200 sec */
  wait_unit();
  wait_unit();
  wait_unit();
  wait_unit();
  wait_unit();
  wait_unit();
  wait_unit();
  wait_unit();
  PORT5 = 1;

  while (x)
  {
    if (code & x) /* wait 4/1200 sec */
    {
      wait_unit();
      wait_unit();
      wait_unit();
      wait_unit();
    }
    else /* wait 1/1200 sec */
      wait_unit();

    PORT5 = 0;
    /* wait 1/1200 sec */
    wait_unit();
    PORT5 = 1;

    /* move to next bit of the hexcode */
    x >>= 1;
  }

  /* set again TE and RE bits */
  S_CR = S_CR | 0x30;

  asm("ANDC #127, ccr"); /* set interrupts */
}

static unsigned char message;  /* message arrived (1 = simple, 2 = sequence, send now, 3 = sequence, on button) */
static int code;               /* code of the simple message */
static int si;                 /* length of sequence message */
static int seq[1024];          /* sequence message codes */
static time_t timing[1024];       /* sequence message timing */


void lnp_integrity_message(const unsigned char *buf, unsigned char ln)
{
  char ctrl;
  if (ln > 2)
  {
    ctrl = (char)buf[0];
    if (ctrl == 0x7E)  /* 7E: sequence start */
    {
      si = 0;
      timing[si] = get_system_up_time();
      message = 0xFF;
    }
    if (ctrl == 0x7D)  /* 7D: sequence data */
    {
      if (message == 0xFF)
      {
        seq[si++] = (int)buf[1] + ((int)buf[2] << 8);
        if (si == 1023) si = 0;
        timing[si] = get_system_up_time();
      }
      else  /* immediate message (was not preceded by sequence start) */
      {
        code = (int)buf[1] + ((int)buf[2] << 8);
        message = 1;
      }
    }
    if (ctrl == 0x7C)  /* 7C: sequence end - send immediatelly */
      message = 2;
    else if (ctrl == 0x7B) /* 7B: sequence end - do not send, wait for PRGM button */
      message = 3;
    else if (ctrl == 0x7A) /* 7A: overwrite time */
      if (si > 0) timing[si] = timing[si - 1] + (int)buf[1] + ((int)buf[2] << 8);
  }
}

int main(int argc, char **argv)
{
  time_t t0;
  int i;

  message = 0;
  lnp_integrity_handler = &lnp_integrity_message;
  lnp_logical_range(1);

  //systime_set_timeslice(255);
  cputs("ready");

  while (RELEASED(dbutton(),BUTTON_VIEW))
  {
    if (PRESSED(dbutton(), BUTTON_PROGRAM))
    {
      msleep(200);
      while (PRESSED(dbutton(), BUTTON_PROGRAM));
      msleep(50);
      message = 2;
    }
    if (message == 1)
    {
      message = 0;
      sapien_send(code);
      cputw(code);
    }
    else if (message == 2)
    {
      cputs("work ");
      message = 0;
      t0 = get_system_up_time() - timing[0];
      i = 1;
      while (i <= si)
      {
	while (get_system_up_time() < timing[i] + t0);
        sapien_send(seq[i-1]);
	cputw(i);
	i++;
      }
      cputs("sent ");
    }
    else if (message == 3)
    {
      cputs("progm");
      message = 0;
    }
    else if (message == 0xFF)
    {
      cputw(si);
    }
  }

  cputs("done");
  sleep(1);
  cls();
  lnp_integrity_handler = LNP_DUMMY_INTEGRITY;

  return 0;
}

